
// Exponential Shadow Map Effect

float4x4	WorldViewProjection;
float4x4	LightViewProjection;
float4x4	ShadowTransform;
texture		ShadowTexture;
float3		LightPosition;
float		NearPlane;
float       FarPlane;
float		LargeConstant = 80;
float		DepthOffset;
float		ShadowThreshold;
float2		InverseResolution;

float3 DiffuseColor = float3(1.00, 1.00, 1.00);
float3 AmbientColor = float3(0.25, 0.30, 0.35);

struct VS_INPUT {
	float4	Position0	: POSITION;
	float3	Normal		: NORMAL0;
	float2	TexCoord	: TEXCOORD0;
};

struct VS_OUTPUT {
	float4	Position0	: POSITION;
	float4	Color		: COLOR0;
	float4	TexCoord	: TEXCOORD0;
	float4	Position1	: TEXCOORD1;
	float3	Normal		: TEXCOORD2;
};

sampler PointSampler = sampler_state {
	texture = <ShadowTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MagFilter = POINT;
	MinFilter = POINT;
};

sampler LinearSampler = sampler_state {
	texture = <ShadowTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MagFilter = LINEAR;
	MinFilter = LINEAR;
};

float Texel(float2 location, float2 offset) {
	return tex2D(LinearSampler, location + (float2(0.5, 0.5) + offset) * InverseResolution).x;
}

float Filter( float2 TexCoord ) {
	float result = Texel(TexCoord, float2( 0,  0)) * 1.0;
	result += Texel(TexCoord, float2( 0, +1));
	result += Texel(TexCoord, float2(-1,  0));
	result += Texel(TexCoord, float2(+1,  0));
	result += Texel(TexCoord, float2( 0, +1));
	result += (
		  Texel(TexCoord, float2(-1, -1))
		+ Texel(TexCoord, float2(+1, -1))
		+ Texel(TexCoord, float2(-1, +1))
		+ Texel(TexCoord, float2(+1, +1))) * 1.0;

	return result / (1.0 + 1.0 * 4 + 1.0 * 4);
}

VS_OUTPUT VSCaster( const VS_INPUT v ) {
	VS_OUTPUT	o = (VS_OUTPUT)0;

	o.Position0 = mul( v.Position0, LightViewProjection );

	o.TexCoord.w = o.Position0.w / FarPlane;
	return o;
}

VS_OUTPUT VSReceiver( const VS_INPUT v ) {
	VS_OUTPUT o = (VS_OUTPUT)0;
	o.Position1 = o.Position0 = mul( v.Position0, WorldViewProjection );
	o.Normal = v.Normal.xyz;
	o.Color.xyzw = 1.0f;
	o.TexCoord = mul( v.Position0, ShadowTransform );
	o.TexCoord.z = mul( v.Position0, LightViewProjection ).w / FarPlane;
	return o;
}

VS_OUTPUT VSTexture( const VS_INPUT v ) {
	VS_OUTPUT o = (VS_OUTPUT)0;
	o.Position0 = v.Position0;
	o.TexCoord.xy = v.TexCoord.xy;
	return o;
}

float4 PSSimpleCaster( const VS_OUTPUT v ) : COLOR {
	return v.TexCoord.w + DepthOffset;
}

float4 PSSimpleReceiver( const VS_OUTPUT v ) : COLOR {
	float4 color = v.Color;
	float3 lightDirection = normalize(LightPosition - v.Position1.xyz / v.Position1.w);
	color.rgb = DiffuseColor * max(0, dot( normalize(v.Normal), lightDirection ));
	color.rgb *= ( tex2Dproj( PointSampler, v.TexCoord ).x >= v.TexCoord.z );
	color.rgb += AmbientColor;
	return color;
}

float4 PSEsmCaster( const VS_OUTPUT v ) : COLOR {
	return exp2((v.TexCoord.w + DepthOffset) * LargeConstant);
}

float4 PSEsmReceiver( const VS_OUTPUT v ) : COLOR {
	float4	color = v.Color;
	float3 lightDirection = normalize(LightPosition - v.Position1.xyz / v.Position1.w);
	float	expZ = tex2Dproj( LinearSampler, v.TexCoord ).x;
	float	expD = exp2(-v.TexCoord.z * LargeConstant);
	color.rgb = DiffuseColor * max(0, dot( normalize(v.Normal), lightDirection ));
	color.rgb *= saturate(expD * expZ - ShadowThreshold);
	color.rgb += AmbientColor;
	return color;
}

float4 PSFilter( const VS_OUTPUT v ) : COLOR {
	return Filter(v.TexCoord);
}

float4 PSSimpleTexture( const VS_OUTPUT v ) : COLOR {
	float4 color;
	color.rgb = tex2D(PointSampler, v.TexCoord.xy).xxx;
	color.a = 1.0f;
	return color;
}

float4 PSEsmTexture( const VS_OUTPUT v ) : COLOR {
	float4 color;
	color.rgb = (log2(tex2D(PointSampler, v.TexCoord.xy)) / LargeConstant).xxx;
	color.a = 1.0f;
	return color;
}

technique ShadowCaster {
	pass Pass0 {
		VertexShader = compile vs_2_0 VSCaster();
		PixelShader = compile ps_2_0 PSSimpleCaster();

		ZEnable = True;
		AlphaBlendEnable = False;
		CullMode = CCW;
	}
}

technique ShadowReceiver {
	pass Pass0 {
		VertexShader = compile vs_2_0 VSReceiver();
		PixelShader = compile ps_2_0 PSSimpleReceiver();

		ZEnable = True;
		AlphaBlendEnable = True;
		CullMode = CCW;
	}
}

technique EsmCaster {
	pass Pass0 {
		VertexShader = compile vs_2_0 VSCaster();
		PixelShader = compile ps_2_0 PSEsmCaster();

		ZEnable = True;
		AlphaBlendEnable = False;
		CullMode = CCW;
	}
}

technique EsmReceiver {
	pass Pass0 {
		VertexShader = compile vs_2_0 VSReceiver();
		PixelShader = compile ps_2_0 PSEsmReceiver();

		ZEnable = True;
		AlphaBlendEnable = True;
		CullMode = CCW;
	}
}

technique Filter {
	pass Pass0 {
		VertexShader = compile vs_2_0 VSTexture();
		PixelShader = compile ps_2_0 PSFilter();

		ZEnable = False;
		AlphaBlendEnable = False;
		CullMode = CCW;
	}
}

technique SimpleTexture {
	pass Pass0 {
		VertexShader = compile vs_2_0 VSTexture();
		PixelShader = compile ps_2_0 PSSimpleTexture();

		ZEnable = False;
		AlphaBlendEnable = False;
		CullMode = CCW;
	}
}

technique EsmTexture {
	pass Pass0 {
		VertexShader = compile vs_2_0 VSTexture();
		PixelShader = compile ps_2_0 PSEsmTexture();

		ZEnable = False;
		AlphaBlendEnable = False;
		CullMode = CCW;
	}
}
