using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace ExponentialShadowMap {

	public class ShadowSettings {
		public bool EsmMode = true;
		public Vector3 CameraPosition = new Vector3(0, 0, -350);
		public float CameraAngleX = 0.8f;
		public float CameraAngleY = 0.2f;
		public float LargeConstant = 89;
		public float ShadowThreshold = 0.0f;
		public float DepthOffset = 0.01f;
		public float LightAngleX = 0.2f;
		public float LightAngleY = -0.5f;
		/// <summary>
		/// tB^L
		/// </summary>
		public bool FilterEnabled = true;
	}

	public class ShadowMap : IDisposable {
		readonly GraphicsDevice graphicsDevice;
		readonly Effect effect;

		const int Width = 2048;
		const int Height = 2048;

		RenderTarget2D shadow;
		RenderTarget2D filter;
		DepthStencilBuffer depth;
		readonly QuadFiller quad;
		readonly ShadowSettings settings = new ShadowSettings();

		public event Action DrawScene;

		public GraphicsDevice GraphicsDevice { get { return graphicsDevice; } }
		public ShadowSettings Settings { get { return settings; } }

		public ShadowMap(GraphicsDevice graphicsDevice, ContentManager content) {
			this.graphicsDevice = graphicsDevice;
			effect = content.Load<Effect>("ShadowMap");
			shadow = new RenderTarget2D(GraphicsDevice, Width, Height, 1, SurfaceFormat.Single, RenderTargetUsage.PreserveContents);
			filter = new RenderTarget2D(GraphicsDevice, Width, Height, 1, SurfaceFormat.Single, RenderTargetUsage.PreserveContents);
			depth = new DepthStencilBuffer(GraphicsDevice, Width, Height, DepthFormat.Depth24, MultiSampleType.None, 0);
			quad = new QuadFiller(GraphicsDevice);
		}
		public void Dispose() {
			shadow.Dispose();
			filter.Dispose();
			depth.Dispose();
			quad.Dispose();
		}

		public virtual void OnDrawScene() {
			if (DrawScene != null) DrawScene();
		}

		void DrawShadowReceiver() {
			GraphicsDevice.Clear(ClearOptions.Target | ClearOptions.DepthBuffer, Color.CornflowerBlue, 1.0f, 0);

			if (settings.EsmMode && settings.FilterEnabled) {
				effect.Parameters["ShadowTexture"].SetValue(filter.GetTexture());
			} else {
				effect.Parameters["ShadowTexture"].SetValue(shadow.GetTexture());
			}

			effect.CurrentTechnique = effect.Techniques[settings.EsmMode ? "EsmReceiver" : "ShadowReceiver"];

			OnDrawScene();
		}

		void DrawShadowCaster() {
			var oldDepth = GraphicsDevice.DepthStencilBuffer;

			GraphicsDevice.SetRenderTarget(0, shadow);
			GraphicsDevice.DepthStencilBuffer = depth;
			GraphicsDevice.Clear(ClearOptions.Target | ClearOptions.DepthBuffer, Color.White, 1.0f, 0);

			 effect.CurrentTechnique = effect.Techniques[settings.EsmMode ? "EsmCaster" : "ShadowCaster"];

			OnDrawScene();

			if (settings.EsmMode && settings.FilterEnabled) {
				GraphicsDevice.SetRenderTarget(0, filter);
				effect.Parameters["ShadowTexture"].SetValue(shadow.GetTexture());
				effect.CurrentTechnique = effect.Techniques["Filter"];
				quad.Fill(effect);
			}

			GraphicsDevice.SetRenderTarget(0, null);
			GraphicsDevice.DepthStencilBuffer = oldDepth;
		}

		public void Draw(Model model) {
			foreach (var mesh in model.Meshes) {
				effect.Begin();
				foreach (var pass in effect.CurrentTechnique.Passes) {
					pass.Begin();
					foreach (var part in mesh.MeshParts) {
						var modelEffect = (BasicEffect)part.Effect;
						effect.Parameters["DiffuseColor"].SetValue(modelEffect.DiffuseColor);
						GraphicsDevice.VertexDeclaration = part.VertexDeclaration;
						GraphicsDevice.Vertices[0].SetSource(mesh.VertexBuffer, 0, part.VertexStride);
						GraphicsDevice.Indices = mesh.IndexBuffer;
						GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, part.BaseVertex,
							part.StartIndex, part.NumVertices, part.StartIndex, part.PrimitiveCount);
					}
					pass.End();
				}
				effect.End();
			}
		}

		void DrawShadowTexture() {
			effect.CurrentTechnique = effect.Techniques[settings.EsmMode ? "EsmTexture" : "SimpleTexture"];
			float size = 0.5f;
			float x0 = -0.8f;
			float y0 = -0.8f;
			float x1 = x0 + size;
			float y1 = y0 + size * GraphicsDevice.Viewport.AspectRatio;
			quad.Fill(effect, x0, y0, x1, y1);
		}

		void UpdateLight(Vector3 lightPosition) {
			Matrix view = Matrix.CreateLookAt(lightPosition, Vector3.Zero, Vector3.Up);

			float nearPlane = 1.0f;
			float farPlane = 500.0f;
			var projection = Matrix.CreatePerspectiveFieldOfView(
				MathHelper.ToRadians(60), 1.0f, nearPlane, farPlane);

			Matrix offset = new Matrix(
				0.5f, 0.0f, 0.0f, 0.0f,
				0.0f, -0.5f, 0.0f, 0.0f,
				0.0f, 0.0f, 1.0f, 0.0f,
				0.5f, 0.5f, 0.0f, 1.0f);

			Matrix lightViewProjection = view * projection;
			Matrix shadow = lightViewProjection * offset;
			effect.Parameters["LightViewProjection"].SetValue(lightViewProjection);
			effect.Parameters["ShadowTransform"].SetValue(shadow);
			effect.Parameters["NearPlane"].SetValue(nearPlane);
			effect.Parameters["FarPlane"].SetValue(farPlane);
		}

		public void UpdateShadow() {
			Matrix world = Matrix.Identity;

			var rotation = Matrix.CreateFromYawPitchRoll(settings.CameraAngleY, settings.CameraAngleX, 0.0f);
			var cameraPosition = Vector3.Transform(settings.CameraPosition, rotation);
			Matrix view = Matrix.CreateLookAt(cameraPosition, Vector3.Zero, Vector3.Up);

			float nearPlane = 1.0f;
			float farPlane = 10000.0f;
			Matrix projection = Matrix.CreatePerspectiveFieldOfView(
				MathHelper.ToRadians(30), GraphicsDevice.Viewport.AspectRatio, nearPlane, farPlane);

			Vector3 lightPosition = new Vector3(0.0f, 200.0f, 200.0f);
			rotation = Matrix.CreateFromYawPitchRoll(settings.LightAngleY, settings.LightAngleX, 0.0f);
			lightPosition = Vector3.Transform(lightPosition, rotation);

			Matrix worldViewProjection = world * view * projection;
			effect.Parameters["WorldViewProjection"].SetValue(worldViewProjection);
			effect.Parameters["LightPosition"].SetValue(lightPosition);
			effect.Parameters["LargeConstant"].SetValue(settings.LargeConstant);
			effect.Parameters["ShadowThreshold"].SetValue(settings.ShadowThreshold);
			effect.Parameters["DepthOffset"].SetValue(settings.DepthOffset);
			effect.Parameters["InverseResolution"].SetValue(new Vector2(1.0f / Width, 1.0f / Height));

			UpdateLight(lightPosition);
		}

		public void Draw() {
			DrawShadowCaster();
			DrawShadowReceiver();
			DrawShadowTexture();
		}
	}

	public class ShadowMapGame : Microsoft.Xna.Framework.Game {
		readonly GraphicsDeviceManager graphics;
		SpriteBatch spriteBatch;
		KeyboardState previous;
		KeyboardState current;

		ShadowMap shadowMap;
		Model model;

		public ShadowMapGame() {
			graphics = new GraphicsDeviceManager(this);
			graphics.PreferMultiSampling = true;
			Content.RootDirectory = "Content";
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);

			model = Content.Load<Model>("Primitives");

			shadowMap = new ShadowMap(GraphicsDevice, Content);

			shadowMap.DrawScene += DrawScene;
		}

		void DrawScene() {
			shadowMap.Draw(model);
		}

		protected override void Update(GameTime gameTime) {
			if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
				this.Exit();

			previous = current;
			current = Keyboard.GetState();

			if (current.IsKeyDown(Keys.Space) && previous.IsKeyUp(Keys.Space)) {
				shadowMap.Settings.EsmMode = !shadowMap.Settings.EsmMode;
			}
			if (current.IsKeyDown(Keys.Left)) {
				shadowMap.Settings.CameraAngleY += 0.1f;
			}
			if (current.IsKeyDown(Keys.Right)) {
				shadowMap.Settings.CameraAngleY -= 0.1f;
			}
			if (current.IsKeyDown(Keys.Up)) {
				shadowMap.Settings.CameraAngleX += 0.1f;
			}
			if (current.IsKeyDown(Keys.Down)) {
				shadowMap.Settings.CameraAngleX -= 0.1f;
			}
			if (current.IsKeyDown(Keys.A)) {
				shadowMap.Settings.LightAngleY += 0.1f;
			}
			if (current.IsKeyDown(Keys.D)) {
				shadowMap.Settings.LightAngleY -= 0.1f;
			}
			if (current.IsKeyDown(Keys.W)) {
				shadowMap.Settings.LightAngleX += 0.1f;
			}
			if (current.IsKeyDown(Keys.S)) {
				shadowMap.Settings.LightAngleX -= 0.1f;
			}

			shadowMap.UpdateShadow();

			base.Update(gameTime);
		}

		protected override void Draw(GameTime gameTime) {
			shadowMap.Draw();
			base.Draw(gameTime);
		}

	}

	class QuadFiller : IDisposable {
		readonly GraphicsDevice graphicsDevice;
		readonly VertexDeclaration declaration;
		public QuadFiller(GraphicsDevice graphicsDevice) {
			this.graphicsDevice = graphicsDevice;
			declaration = new VertexDeclaration(graphicsDevice, VertexPositionTexture.VertexElements);
		}
		readonly VertexPositionTexture[] vertices = new VertexPositionTexture[] {
			new VertexPositionTexture(new Vector3(-1, +1, 0), new Vector2(0, 0)),
			new VertexPositionTexture(new Vector3(+1, +1, 0), new Vector2(1, 0)),
			new VertexPositionTexture(new Vector3(-1, -1, 0), new Vector2(0, 1)),
			new VertexPositionTexture(new Vector3(+1, -1, 0), new Vector2(1, 1)),
		};
		public void Dispose() {
			declaration.Dispose();
		}
		public void Fill(Effect effect) {
			effect.Begin();
			foreach (var pass in effect.CurrentTechnique.Passes) {
				pass.Begin();
				Fill();
				pass.End();
			}
			effect.End();
		}
		public void Fill() {
			graphicsDevice.VertexDeclaration = declaration;
			graphicsDevice.DrawUserPrimitives(PrimitiveType.TriangleStrip, vertices, 0, 2);
		}
		public void Fill(Effect effect, float x0, float y0, float x1, float y1) {
			effect.Begin();
			foreach (var pass in effect.CurrentTechnique.Passes) {
				pass.Begin();
				Fill(x0, y0, x1, y1);
				pass.End();
			}
			effect.End();
		}
		public void Fill(float x0, float y0, float x1, float y1) {
			VertexPositionTexture[] vertices = new VertexPositionTexture[] {
				new VertexPositionTexture(new Vector3(x0, y1, 0), new Vector2(0, 0)),
				new VertexPositionTexture(new Vector3(x1, y1, 0), new Vector2(1, 0)),
				new VertexPositionTexture(new Vector3(x0, y0, 0), new Vector2(0, 1)),
				new VertexPositionTexture(new Vector3(x1, y0, 0), new Vector2(1, 1)),
			};
			graphicsDevice.VertexDeclaration = declaration;
			graphicsDevice.DrawUserPrimitives(PrimitiveType.TriangleStrip, vertices, 0, 2);
		}
	}

}
